from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, List

from ai21.clients.common.beta.assistant.plans import BasePlans
from ai21.clients.common.beta.assistant.routes import BaseRoutes
from ai21.models.assistant.assistant import Optimization, Tool, ToolResources
from ai21.models.responses.assistant_response import AssistantResponse, ListAssistant
from ai21.types import NotGiven, NOT_GIVEN
from ai21.utils.typing import remove_not_given


class BaseAssistants(ABC):
    _module_name = "assistants"
    plans: BasePlans
    routes: BaseRoutes

    @abstractmethod
    def create(
        self,
        name: str,
        *,
        description: str | NotGiven = NOT_GIVEN,
        optimization: Optimization | NotGiven = NOT_GIVEN,
        models: List[str] | NotGiven = NOT_GIVEN,
        tools: List[Tool] | NotGiven = NOT_GIVEN,
        tool_resources: ToolResources | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> AssistantResponse:
        pass

    def _create_body(
        self,
        *,
        name: str,
        description: str | NotGiven,
        optimization: str | NotGiven,
        models: List[str] | NotGiven,
        tools: List[str] | NotGiven,
        tool_resources: dict | NotGiven,
        **kwargs,
    ) -> Dict[str, Any]:
        return remove_not_given(
            {
                "name": name,
                "description": description,
                "optimization": optimization,
                "models": models,
                "tools": tools,
                "tool_resources": tool_resources,
                **kwargs,
            }
        )

    @abstractmethod
    def list(self) -> ListAssistant:
        pass

    @abstractmethod
    def retrieve(self, assistant_id: str) -> AssistantResponse:
        pass

    @abstractmethod
    def modify(
        self,
        assistant_id: str,
        *,
        name: str | NotGiven = NOT_GIVEN,
        description: str | NotGiven = NOT_GIVEN,
        optimization: Optimization | NotGiven = NOT_GIVEN,
        is_archived: bool | NotGiven = NOT_GIVEN,
        models: List[str] | NotGiven = NOT_GIVEN,
        tools: List[Tool] | NotGiven = NOT_GIVEN,
        tool_resources: ToolResources | NotGiven = NOT_GIVEN,
    ) -> AssistantResponse:
        pass



from typing import List

from ai21.models.ai21_base_model import AI21BaseModel


class TopTokenData(AI21BaseModel):
    token: str
    logprob: float


class LogprobsData(AI21BaseModel):
    token: str
    logprob: float
    top_logprobs: List[TopTokenData]


class Logprobs(AI21BaseModel):
    content: LogprobsData



from datetime import datetime
from typing import Optional, List, Literal

from ai21.models.ai21_base_model import AI21BaseModel
from ai21.models.assistant.assistant import ToolResources


class AssistantResponse(AI21BaseModel):
    id: str
    created_at: datetime
    updated_at: datetime
    object: Literal["assistant"] = "assistant"
    name: str
    description: Optional[str] = None
    optimization: str
    organization_id: str
    user_id: str
    avatar: Optional[str] = None
    is_archived: bool = False
    models: Optional[List[str]] = None
    tools: Optional[List[str]] = None
    tool_resources: Optional[ToolResources] = None


class ListAssistant(AI21BaseModel):
    results: List[AssistantResponse]



from __future__ import annotations

from typing import List

from ai21.clients.common.beta.assistant.routes import BaseRoutes
from ai21.clients.studio.resources.studio_resource import (
    AsyncStudioResource,
    StudioResource,
)
from ai21.models.responses.route_response import RouteResponse, ListRouteResponse
from ai21.types import NotGiven, NOT_GIVEN


class AssistantRoutes(StudioResource, BaseRoutes):
    def create(
        self,
        *,
        assistant_id: str,
        plan_id: str,
        name: str,
        description: str,
        examples: List[str],
        **kwargs,
    ) -> RouteResponse:
        body = self._create_body(
            plan_id=plan_id,
            name=name,
            description=description,
            examples=examples,
            **kwargs,
        )

        return self._post(path=f"/assistants/{assistant_id}/{self._module_name}", body=body, response_cls=RouteResponse)

    def list(
        self,
        *,
        assistant_id: str,
        name: str | NotGiven = NOT_GIVEN,
    ) -> ListRouteResponse:
        params = self._create_body(name=name)

        return self._get(
            path=f"/assistants/{assistant_id}/{self._module_name}", params=params, response_cls=ListRouteResponse
        )

    def retrieve(
        self,
        *,
        assistant_id: str,
        route_id: str,
    ) -> RouteResponse:
        return self._get(path=f"/assistants/{assistant_id}/{self._module_name}/{route_id}", response_cls=RouteResponse)

    def modify(
        self,
        *,
        assistant_id: str,
        route_id: str,
        description: str | NotGiven = NOT_GIVEN,
        examples: List[str] | NotGiven = NOT_GIVEN,
    ) -> RouteResponse:
        body = self._create_body(
            description=description,
            examples=examples,
        )

        return self._patch(
            path=f"/assistants/{assistant_id}/{self._module_name}/{route_id}", body=body, response_cls=RouteResponse
        )

    def delete(
        self,
        *,
        assistant_id: str,
        route_id: str,
    ):
        return self._delete(path=f"/assistants/{assistant_id}/{self._module_name}/{route_id}")


class AsyncAssistantRoutes(AsyncStudioResource, BaseRoutes):
    async def create(
        self,
        *,
        assistant_id: str,
        plan_id: str,
        name: str,
        description: str,
        examples: List[str],
        **kwargs,
    ) -> RouteResponse:
        body = self._create_body(
            plan_id=plan_id,
            name=name,
            description=description,
            examples=examples,
            **kwargs,
        )

        return await self._post(
            path=f"/assistants/{assistant_id}/{self._module_name}", body=body, response_cls=RouteResponse
        )

    async def list(
        self,
        *,
        assistant_id: str,
        name: str | NotGiven = NOT_GIVEN,
    ) -> ListRouteResponse:
        params = self._create_body(name=name)

        return await self._get(
            path=f"/assistants/{assistant_id}/{self._module_name}", params=params, response_cls=ListRouteResponse
        )

    async def retrieve(
        self,
        *,
        assistant_id: str,
        route_id: str,
    ) -> RouteResponse:
        return await self._get(
            path=f"/assistants/{assistant_id}/{self._module_name}/{route_id}", response_cls=RouteResponse
        )

    async def modify(
        self,
        *,
        assistant_id: str,
        route_id: str,
        description: str | NotGiven = NOT_GIVEN,
        examples: List[str] | NotGiven = NOT_GIVEN,
    ) -> RouteResponse:
        body = self._create_body(
            description=description,
            examples=examples,
        )

        return await self._patch(
            path=f"/assistants/{assistant_id}/{self._module_name}/{route_id}", body=body, response_cls=RouteResponse
        )

    async def delete(
        self,
        *,
        assistant_id: str,
        route_id: str,
    ):
        return await self._delete(path=f"/assistants/{assistant_id}/{self._module_name}/{route_id}")



from typing import Optional, Any, Dict

import httpx

from ai21.ai21_env_config import _AI21EnvConfig, AI21EnvConfig
from ai21.clients.studio.client_url_parser import create_client_url
from ai21.clients.studio.resources.beta.async_beta import AsyncBeta
from ai21.clients.studio.resources.studio_chat import AsyncStudioChat
from ai21.clients.studio.resources.studio_library import AsyncStudioLibrary
from ai21.http_client.async_http_client import AsyncAI21HTTPClient


class AsyncAI21Client(AsyncAI21HTTPClient):
    """
    This class would be sending requests to our REST API using http requests asynchronously
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        api_host: Optional[str] = None,
        headers: Optional[Dict[str, Any]] = None,
        timeout_sec: Optional[float] = None,
        num_retries: Optional[int] = None,
        via: Optional[str] = None,
        http_client: Optional[httpx.AsyncClient] = None,
        env_config: _AI21EnvConfig = AI21EnvConfig,
        **kwargs,
    ):
        base_url = create_client_url(api_host or env_config.api_host)

        super().__init__(
            api_key=api_key or env_config.api_key,
            base_url=base_url,
            headers=headers,
            timeout_sec=timeout_sec or env_config.timeout_sec,
            num_retries=num_retries or env_config.num_retries,
            via=via,
            client=http_client,
        )

        self.chat: AsyncStudioChat = AsyncStudioChat(self)
        self.library = AsyncStudioLibrary(self)
        self.beta = AsyncBeta(self)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List

from ai21.models.assistant.assistant import Optimization
from ai21.models.assistant.run import ToolOutput
from ai21.models.responses.run_response import RunResponse
from ai21.types import NOT_GIVEN, NotGiven
from ai21.utils.typing import remove_not_given


class BaseRuns(ABC):
    _module_name = "runs"

    @abstractmethod
    def create(
        self,
        *,
        thread_id: str,
        assistant_id: str,
        description: str | NotGiven = NOT_GIVEN,
        optimization: Optimization | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> RunResponse:
        pass

    def _create_body(
        self,
        *,
        thread_id: str,
        assistant_id: str,
        description: str | NotGiven,
        optimization: str | NotGiven,
        **kwargs,
    ) -> dict:
        return remove_not_given(
            {
                "thread_id": thread_id,
                "assistant_id": assistant_id,
                "description": description,
                "optimization": optimization,
                **kwargs,
            }
        )

    @abstractmethod
    def retrieve(
        self,
        *,
        thread_id: str,
        run_id: str,
    ) -> RunResponse:
        pass

    @abstractmethod
    def cancel(
        self,
        *,
        thread_id: str,
        run_id: str,
    ) -> RunResponse:
        pass

    @abstractmethod
    def submit_tool_outputs(self, *, thread_id: str, run_id: str, tool_outputs: List[ToolOutput]) -> RunResponse:
        pass



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List

from ai21.clients.common.beta.assistant.messages import BaseMessages
from ai21.models.assistant.message import Message
from ai21.models.responses.thread_response import ThreadResponse


class BaseThreads(ABC):
    _module_name = "threads"
    messages: BaseMessages

    @abstractmethod
    def create(
        self,
        messages: List[Message],
        **kwargs,
    ) -> ThreadResponse:
        pass

    @abstractmethod
    def retrieve(self, thread_id: str) -> ThreadResponse:
        pass



from __future__ import annotations

from typing import List

from ai21.clients.common.beta.assistant.assistants import BaseAssistants
from ai21.clients.studio.resources.beta.assistant.assistants_plans import AssistantPlans, AsyncAssistantPlans
from ai21.clients.studio.resources.beta.assistant.assistant_routes import AssistantRoutes, AsyncAssistantRoutes
from ai21.clients.studio.resources.studio_resource import (
    AsyncStudioResource,
    StudioResource,
)
from ai21.http_client.async_http_client import AsyncAI21HTTPClient
from ai21.http_client.http_client import AI21HTTPClient
from ai21.models.assistant.assistant import Tool, ToolResources
from ai21.models.responses.assistant_response import AssistantResponse, ListAssistant
from ai21.types import NotGiven, NOT_GIVEN


class Assistants(StudioResource, BaseAssistants):
    def __init__(self, client: AI21HTTPClient):
        super().__init__(client)

        self.plans = AssistantPlans(client)
        self.routes = AssistantRoutes(client)

    def create(
        self,
        name: str,
        *,
        description: str | NotGiven = NOT_GIVEN,
        optimization: str | NotGiven = NOT_GIVEN,
        models: List[str] | NotGiven = NOT_GIVEN,
        tools: List[Tool] | NotGiven = NOT_GIVEN,
        tool_resources: ToolResources | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> AssistantResponse:
        body = self._create_body(
            name=name,
            description=description,
            optimization=optimization,
            models=models,
            tools=tools,
            tool_resources=tool_resources,
            **kwargs,
        )

        return self._post(path=f"/{self._module_name}", body=body, response_cls=AssistantResponse)

    def retrieve(self, assistant_id: str) -> AssistantResponse:
        return self._get(path=f"/{self._module_name}/{assistant_id}", response_cls=AssistantResponse)

    def list(self) -> ListAssistant:
        return self._get(path=f"/{self._module_name}", response_cls=ListAssistant)

    def modify(
        self,
        assistant_id: str,
        *,
        name: str | NotGiven = NOT_GIVEN,
        description: str | NotGiven = NOT_GIVEN,
        optimization: str | NotGiven = NOT_GIVEN,
        is_archived: bool | NotGiven = NOT_GIVEN,
        models: List[str] | NotGiven = NOT_GIVEN,
        tools: List[Tool] | NotGiven = NOT_GIVEN,
        tool_resources: ToolResources | NotGiven = NOT_GIVEN,
    ) -> AssistantResponse:
        body = self._create_body(
            name=name,
            description=description,
            optimization=optimization,
            is_archived=is_archived,
            models=models,
            tools=tools,
            tool_resources=tool_resources,
        )

        return self._patch(path=f"/{self._module_name}/{assistant_id}", body=body, response_cls=AssistantResponse)


class AsyncAssistants(AsyncStudioResource, BaseAssistants):
    def __init__(self, client: AsyncAI21HTTPClient):
        super().__init__(client)

        self.plans = AsyncAssistantPlans(client)
        self.routes = AsyncAssistantRoutes(client)

    async def create(
        self,
        name: str,
        *,
        description: str | NotGiven = NOT_GIVEN,
        optimization: str | NotGiven = NOT_GIVEN,
        models: List[str] | NotGiven = NOT_GIVEN,
        tools: List[Tool] | NotGiven = NOT_GIVEN,
        tool_resources: ToolResources | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> AssistantResponse:
        body = self._create_body(
            name=name,
            description=description,
            optimization=optimization,
            models=models,
            tools=tools,
            tool_resources=tool_resources,
            **kwargs,
        )

        return await self._post(path=f"/{self._module_name}", body=body, response_cls=AssistantResponse)

    async def retrieve(self, assistant_id: str) -> AssistantResponse:
        return await self._get(path=f"/{self._module_name}/{assistant_id}", response_cls=AssistantResponse)

    async def list(self) -> ListAssistant:
        return await self._get(path=f"/{self._module_name}", response_cls=ListAssistant)

    async def modify(
        self,
        assistant_id: str,
        *,
        name: str | NotGiven = NOT_GIVEN,
        description: str | NotGiven = NOT_GIVEN,
        optimization: str | NotGiven = NOT_GIVEN,
        is_archived: bool | NotGiven = NOT_GIVEN,
        models: List[str] | NotGiven = NOT_GIVEN,
        tools: List[Tool] | NotGiven = NOT_GIVEN,
        tool_resources: ToolResources | NotGiven = NOT_GIVEN,
    ) -> AssistantResponse:
        body = self._create_body(
            name=name,
            description=description,
            optimization=optimization,
            is_archived=is_archived,
            models=models,
            tools=tools,
            tool_resources=tool_resources,
        )

        return await self._patch(path=f"/{self._module_name}/{assistant_id}", body=body, response_cls=AssistantResponse)



from typing import Literal, Any, List

from typing_extensions import TypedDict


RunStatus = Literal[
    "cancelled",
    "cancelling",
    "completed",
    "expired",
    "failed",
    "incomplete",
    "in_progress",
    "queued",
    "requires_action",
]


class ToolOutput(TypedDict):
    tool_call_id: str
    output: Any


class Function(TypedDict):
    name: str
    arguments: Any


class ToolCall(TypedDict):
    type: Literal["function"]
    id: str
    function: Function


class SubmitToolCallOutputs(TypedDict):
    tool_calls: List[ToolOutput]


class RequiredAction(TypedDict):
    type: Literal["submit_tool_outputs"]
    submit_tool_outputs: SubmitToolCallOutputs



from __future__ import annotations

from typing import List, Dict, Optional

from ai21.clients.common.completion_base import Completion
from ai21.clients.studio.resources.studio_resource import StudioResource, AsyncStudioResource
from ai21.models import Penalty, CompletionsResponse
from ai21.types import NOT_GIVEN, NotGiven


class StudioCompletion(StudioResource, Completion):
    def create(
        self,
        prompt: str,
        *,
        model: Optional[str] = None,
        max_tokens: int | NotGiven = NOT_GIVEN,
        num_results: int | NotGiven = NOT_GIVEN,
        min_tokens: int | NotGiven = NOT_GIVEN,
        temperature: float | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven = NOT_GIVEN,
        top_k_return: int | NotGiven = NOT_GIVEN,
        stop_sequences: List[str] | NotGiven = NOT_GIVEN,
        frequency_penalty: Penalty | NotGiven = NOT_GIVEN,
        presence_penalty: Penalty | NotGiven = NOT_GIVEN,
        count_penalty: Penalty | NotGiven = NOT_GIVEN,
        epoch: int | NotGiven = NOT_GIVEN,
        logit_bias: Dict[str, float] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> CompletionsResponse:
        model = self._get_model(model=model, model_id=kwargs.pop("model_id", None))
        path = self._get_completion_path(model=model)
        body = self._create_body(
            model=model,
            prompt=prompt,
            max_tokens=max_tokens,
            num_results=num_results,
            min_tokens=min_tokens,
            temperature=temperature,
            top_p=top_p,
            top_k_return=top_k_return,
            stop_sequences=stop_sequences,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            count_penalty=count_penalty,
            epoch=epoch,
            logit_bias=logit_bias,
            **kwargs,
        )
        return self._post(path=path, body=body, response_cls=CompletionsResponse)


class AsyncStudioCompletion(AsyncStudioResource, Completion):
    async def create(
        self,
        prompt: str,
        *,
        model: Optional[str] = None,
        max_tokens: int | NotGiven = NOT_GIVEN,
        num_results: int | NotGiven = NOT_GIVEN,
        min_tokens: int | NotGiven = NOT_GIVEN,
        temperature: float | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven = NOT_GIVEN,
        top_k_return: int | NotGiven = NOT_GIVEN,
        stop_sequences: List[str] | NotGiven = NOT_GIVEN,
        frequency_penalty: Penalty | NotGiven = NOT_GIVEN,
        presence_penalty: Penalty | NotGiven = NOT_GIVEN,
        count_penalty: Penalty | NotGiven = NOT_GIVEN,
        epoch: int | NotGiven = NOT_GIVEN,
        logit_bias: Dict[str, float] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> CompletionsResponse:
        model = self._get_model(model=model, model_id=kwargs.pop("model_id", None))
        path = self._get_completion_path(model=model)
        body = self._create_body(
            model=model,
            prompt=prompt,
            max_tokens=max_tokens,
            num_results=num_results,
            min_tokens=min_tokens,
            temperature=temperature,
            top_p=top_p,
            top_k_return=top_k_return,
            stop_sequences=stop_sequences,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            count_penalty=count_penalty,
            epoch=epoch,
            logit_bias=logit_bias,
            **kwargs,
        )

        return await self._post(path=path, body=body, response_cls=CompletionsResponse)



from typing import List, Optional, Union, Any, Dict

from pydantic import Field

from ai21.models.ai21_base_model import AI21BaseModel


class Prompt(AI21BaseModel):
    text: Optional[str]
    tokens: Optional[List[Dict[str, Any]]] = None


class CompletionData(AI21BaseModel):
    text: Optional[str]
    tokens: Optional[List[Dict[str, Any]]] = None


class CompletionFinishReason(AI21BaseModel):
    reason: Optional[str] = None
    length: Optional[int] = None


class Completion(AI21BaseModel):
    data: CompletionData
    finish_reason: Optional[CompletionFinishReason] = Field(default=None, alias="finishReason")


class CompletionsResponse(AI21BaseModel):
    id: Union[int, str]
    prompt: Prompt
    completions: List[Completion]



from __future__ import annotations

from typing import List

from ai21.clients.common.beta.assistant.threads import BaseThreads
from ai21.clients.studio.resources.beta.assistant.thread_messages import ThreadMessages, AsyncThreadMessages
from ai21.clients.studio.resources.beta.assistant.thread_runs import AsyncThreadRuns, ThreadRuns
from ai21.clients.studio.resources.studio_resource import StudioResource, AsyncStudioResource
from ai21.http_client.async_http_client import AsyncAI21HTTPClient
from ai21.http_client.http_client import AI21HTTPClient
from ai21.models.assistant.message import Message
from ai21.models.responses.thread_response import ThreadResponse


class Threads(StudioResource, BaseThreads):
    def __init__(self, client: AI21HTTPClient):
        super().__init__(client)

        self.messages = ThreadMessages(client)
        self.runs = ThreadRuns(client)

    def create(
        self,
        messages: List[Message],
        **kwargs,
    ) -> ThreadResponse:
        body = dict(messages=messages)

        return self._post(path=f"/{self._module_name}", body=body, response_cls=ThreadResponse)

    def retrieve(self, thread_id: str) -> ThreadResponse:
        return self._get(path=f"/{self._module_name}/{thread_id}", response_cls=ThreadResponse)


class AsyncThreads(AsyncStudioResource, BaseThreads):
    def __init__(self, client: AsyncAI21HTTPClient):
        super().__init__(client)

        self.messages = AsyncThreadMessages(client)
        self.runs = AsyncThreadRuns(client)

    async def create(
        self,
        messages: List[Message],
        **kwargs,
    ) -> ThreadResponse:
        body = dict(messages=messages)

        return await self._post(path=f"/{self._module_name}", body=body, response_cls=ThreadResponse)

    async def retrieve(self, thread_id: str) -> ThreadResponse:
        return await self._get(path=f"/{self._module_name}/{thread_id}", response_cls=ThreadResponse)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, List

from ai21.models.responses.route_response import RouteResponse, ListRouteResponse
from ai21.types import NotGiven, NOT_GIVEN
from ai21.utils.typing import remove_not_given


class BaseRoutes(ABC):
    _module_name = "routes"

    @abstractmethod
    def create(
        self,
        *,
        assistant_id: str,
        plan_id: str,
        name: str,
        description: str,
        examples: List[str],
        **kwargs,
    ) -> RouteResponse:
        pass

    def _create_body(
        self,
        *,
        plan_id: str | NotGiven = NOT_GIVEN,
        name: str | NotGiven = NOT_GIVEN,
        description: str | NotGiven = NOT_GIVEN,
        examples: List[str] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> Dict[str, Any]:
        return remove_not_given(
            {
                "plan_id": plan_id,
                "name": name,
                "description": description,
                "examples": examples,
                **kwargs,
            }
        )

    @abstractmethod
    def retrieve(
        self,
        *,
        assistant_id: str,
        route_id: str,
    ) -> RouteResponse:
        pass

    @abstractmethod
    def list(
        self,
        *,
        assistant_id: str,
        name: str | NotGiven = NotGiven,
    ) -> ListRouteResponse:
        pass

    @abstractmethod
    def modify(
        self,
        *,
        assistant_id: str,
        route_id: str,
        description: str | NotGiven = NOT_GIVEN,
        examples: List[str] | NotGiven = NOT_GIVEN,
    ) -> RouteResponse:
        pass

    @abstractmethod
    def delete(self, *, assistant_id: str, route_id: str):
        pass



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from typing import TypeVar, Iterator, AsyncIterator, Optional

import httpx

from ai21.errors import StreamingDecodeError

_T = TypeVar("_T")
_SSE_DATA_PREFIX = "data: "
_SSE_DONE_MSG = "[DONE]"


def get_stream_message(chunk: str, cast_to: type[_T]) -> Iterator[_T] | AsyncIterator[_T]:
    try:
        chunk = json.loads(chunk)
        if hasattr(cast_to, "from_dict"):
            return cast_to.from_dict(chunk)
        else:
            return cast_to(**chunk)
    except json.JSONDecodeError:
        raise StreamingDecodeError(chunk)


class _SSEDecoderBase(ABC):
    @abstractmethod
    def iter(self, response: httpx.Response) -> Iterator[str]:
        pass

    @abstractmethod
    async def aiter(self, response: httpx.Response) -> AsyncIterator[str]:
        pass


class _SSEDecoder(_SSEDecoderBase):
    def iter(self, response: httpx.Response):
        for line in response.iter_lines():
            line = line.strip()
            decoded_line = self._decode(line)

            if decoded_line is not None:
                yield decoded_line

    async def aiter(self, response: httpx.Response):
        async for line in response.aiter_lines():
            line = line.strip()
            decoded_line = self._decode(line)

            if decoded_line is not None:
                yield decoded_line

    def _decode(self, line: str) -> Optional[str]:
        if not line:
            return None

        if line.startswith(_SSE_DATA_PREFIX):
            return line.strip(_SSE_DATA_PREFIX)

        raise StreamingDecodeError(f"Invalid SSE line: {line}")



from typing import Optional, Literal

from typing_extensions import TypedDict

Optimization = Literal["cost", "latency"]
Tool = Literal["rag", "web_search", "plan_approval"]


class ToolResources(TypedDict, total=False):
    rag: Optional[dict]
    web_search: Optional[dict]
    plan_approval: Optional[dict]



from datetime import datetime
from typing import List, Literal

from ai21.models.ai21_base_model import AI21BaseModel


class ThreadResponse(AI21BaseModel):
    id: str
    created_at: datetime
    updated_at: datetime
    object: Literal["thread"] = "thread"


class ListThread(AI21BaseModel):
    results: List[ThreadResponse]



from datetime import datetime
from typing import Literal, Optional, List

from ai21.models.ai21_base_model import AI21BaseModel
from ai21.models.assistant.message import ThreadMessageRole, MessageContentText


class MessageResponse(AI21BaseModel):
    id: str
    created_at: datetime
    updated_at: datetime
    object: Literal["message"] = "message"
    role: ThreadMessageRole
    content: MessageContentText
    run_id: Optional[str] = None
    assistant_id: Optional[str] = None


class ListMessageResponse(AI21BaseModel):
    results: List[MessageResponse]



from __future__ import annotations

from abc import ABC, abstractmethod

from ai21.models.assistant.message import ThreadMessageRole, MessageContentText
from ai21.models.responses.message_response import MessageResponse, ListMessageResponse


class BaseMessages(ABC):
    _module_name = "messages"

    @abstractmethod
    def create(
        self,
        thread_id: str,
        *,
        role: ThreadMessageRole,
        content: MessageContentText,
        **kwargs,
    ) -> MessageResponse:
        pass

    @abstractmethod
    def list(self, thread_id: str) -> ListMessageResponse:
        pass



from __future__ import annotations

import logging
import os
from dataclasses import dataclass
from typing import Optional

from ai21.constants import DEFAULT_API_VERSION, STUDIO_HOST

# Constants for environment variable keys
_ENV_API_KEY = "AI21_API_KEY"
_ENV_API_VERSION = "AI21_API_VERSION"
_ENV_API_HOST = "AI21_API_HOST"
_ENV_TIMEOUT_SEC = "AI21_TIMEOUT_SEC"
_ENV_NUM_RETRIES = "AI21_NUM_RETRIES"
_ENV_AWS_REGION = "AI21_AWS_REGION"
_ENV_LOG_LEVEL = "AI21_LOG_LEVEL"

_logger = logging.getLogger(__name__)


@dataclass
class _AI21EnvConfig:
    _api_key: Optional[str] = None
    _api_version: str = DEFAULT_API_VERSION
    _api_host: str = STUDIO_HOST
    _timeout_sec: Optional[int] = None
    _num_retries: Optional[int] = None
    _aws_region: Optional[str] = None
    _log_level: Optional[str] = None

    @classmethod
    def from_env(cls) -> _AI21EnvConfig:
        return cls(
            _api_key=os.getenv(_ENV_API_KEY),
            _api_version=os.getenv(_ENV_API_VERSION, DEFAULT_API_VERSION),
            _api_host=os.getenv(_ENV_API_HOST, STUDIO_HOST),
            _timeout_sec=os.getenv(_ENV_TIMEOUT_SEC),
            _num_retries=os.getenv(_ENV_NUM_RETRIES),
            _aws_region=os.getenv(_ENV_AWS_REGION, "us-east-1"),
            _log_level=os.getenv(_ENV_LOG_LEVEL, "info"),
        )

    @property
    def api_key(self) -> str:
        self._api_key = os.getenv(_ENV_API_KEY, self._api_key)
        return self._api_key

    @property
    def api_version(self) -> str:
        self._api_version = os.getenv(_ENV_API_VERSION, self._api_version)
        return self._api_version

    @property
    def api_host(self) -> str:
        self._api_host = os.getenv(_ENV_API_HOST, self._api_host)
        return self._api_host

    @property
    def timeout_sec(self) -> Optional[int]:
        timeout_str = os.getenv(_ENV_TIMEOUT_SEC)

        if timeout_str is not None:
            self._timeout_sec = int(timeout_str)

        return self._timeout_sec

    @property
    def num_retries(self) -> Optional[int]:
        retries_str = os.getenv(_ENV_NUM_RETRIES)

        if retries_str is not None:
            self._num_retries = int(retries_str)

        return self._num_retries

    @property
    def aws_region(self) -> Optional[str]:
        self._aws_region = os.getenv(_ENV_AWS_REGION, self._aws_region)
        return self._aws_region

    @property
    def log_level(self) -> Optional[str]:
        self._log_level = os.getenv(_ENV_LOG_LEVEL, self._log_level)
        return self._log_level

    def log(self, with_secrets: bool = False) -> None:
        env_vars = {
            _ENV_API_VERSION: self.api_version,
            _ENV_API_HOST: self.api_host,
            _ENV_TIMEOUT_SEC: self.timeout_sec,
            _ENV_NUM_RETRIES: self.num_retries,
            _ENV_AWS_REGION: self.aws_region,
            _ENV_LOG_LEVEL: self.log_level,
        }

        if with_secrets:
            env_vars[_ENV_API_KEY] = self.api_key

        _logger.debug(f"AI21 environment configuration: {env_vars}")


AI21EnvConfig = _AI21EnvConfig.from_env()



from __future__ import annotations

from typing import Optional, List

from ai21.clients.studio.resources.studio_resource import StudioResource, AsyncStudioResource
from ai21.http_client.async_http_client import AsyncAI21HTTPClient
from ai21.http_client.http_client import AI21HTTPClient
from ai21.models import FileResponse
from ai21.types import NotGiven, NOT_GIVEN
from ai21.utils.typing import remove_not_given


class StudioLibrary(StudioResource):
    _module_name = "library/files"

    def __init__(self, client: AI21HTTPClient):
        super().__init__(client)
        self.files = LibraryFiles(client)


class LibraryFiles(StudioResource):
    _module_name = "library/files"

    def create(
        self,
        file_path: str,
        *,
        path: Optional[str] | NotGiven = NOT_GIVEN,
        labels: Optional[List[str]] | NotGiven = NOT_GIVEN,
        public_url: Optional[str] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> str:
        files = {"file": open(file_path, "rb")}
        body = remove_not_given({"path": path, "labels": labels, "publicUrl": public_url, **kwargs})

        raw_response = self._post(path=f"/{self._module_name}", files=files, body=body, response_cls=dict)

        return raw_response["fileId"]

    def get(self, file_id: str) -> FileResponse:
        return self._get(path=f"/{self._module_name}/{file_id}", response_cls=FileResponse)

    def list(
        self,
        *,
        offset: Optional[int] | NotGiven = NOT_GIVEN,
        limit: Optional[int] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> List[FileResponse]:
        params = remove_not_given({"offset": offset, "limit": limit})

        return self._get(path=f"/{self._module_name}", params=params, response_cls=List[FileResponse])

    def update(
        self,
        file_id: str,
        *,
        public_url: Optional[str] | NotGiven = NOT_GIVEN,
        labels: Optional[List[str]] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> None:
        body = remove_not_given(
            {
                "publicUrl": public_url,
                "labels": labels,
                **kwargs,
            }
        )
        self._put(path=f"/{self._module_name}/{file_id}", body=body)

    def delete(self, file_id: str) -> None:
        self._delete(path=f"/{self._module_name}/{file_id}")


class AsyncStudioLibrary(AsyncStudioResource):
    _module_name = "library/files"

    def __init__(self, client: AsyncAI21HTTPClient):
        super().__init__(client)
        self.files = AsyncLibraryFiles(client)


class AsyncLibraryFiles(AsyncStudioResource):
    _module_name = "library/files"

    async def create(
        self,
        file_path: str,
        *,
        path: Optional[str] | NotGiven = NOT_GIVEN,
        labels: Optional[List[str]] | NotGiven = NOT_GIVEN,
        public_url: Optional[str] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> str:
        files = {"file": open(file_path, "rb")}
        body = remove_not_given({"path": path, "labels": labels, "publicUrl": public_url, **kwargs})

        raw_response = await self._post(path=f"/{self._module_name}", files=files, body=body, response_cls=dict)

        return raw_response["fileId"]

    async def get(self, file_id: str) -> FileResponse:
        return await self._get(path=f"/{self._module_name}/{file_id}", response_cls=FileResponse)

    async def list(
        self,
        *,
        offset: Optional[int] | NotGiven = NOT_GIVEN,
        limit: Optional[int] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> List[FileResponse]:
        params = remove_not_given({"offset": offset, "limit": limit})

        return await self._get(path=f"/{self._module_name}", params=params, response_cls=List[FileResponse])

    async def update(
        self,
        file_id: str,
        *,
        public_url: Optional[str] | NotGiven = NOT_GIVEN,
        labels: Optional[List[str]] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> None:
        body = remove_not_given(
            {
                "publicUrl": public_url,
                "labels": labels,
                **kwargs,
            }
        )
        await self._put(path=f"/{self._module_name}/{file_id}", body=body)

    async def delete(self, file_id: str) -> None:
        await self._delete(path=f"/{self._module_name}/{file_id}")



from __future__ import annotations

from typing import List, Optional

from ai21.clients.common.chat_base import Chat
from ai21.clients.studio.resources.chat import ChatCompletions, AsyncChatCompletions
from ai21.models.chat import ChatMessage as JambaChatMessage
from ai21.clients.studio.resources.studio_resource import StudioResource, AsyncStudioResource
from ai21.models import ChatMessage, Penalty, ChatResponse
from ai21.types import NotGiven, NOT_GIVEN
from ai21.clients.studio.resources.constants import (
    CHAT_DEFAULT_NUM_RESULTS,
    CHAT_DEFAULT_TEMPERATURE,
    CHAT_DEFAULT_MAX_TOKENS,
    CHAT_DEFAULT_MIN_TOKENS,
    CHAT_DEFAULT_TOP_P,
    CHAT_DEFAULT_TOP_K_RETURN,
)


class StudioChat(StudioResource, Chat):
    def create(
        self,
        model: str,
        messages: List[ChatMessage],
        system: str,
        *,
        num_results: Optional[int] = CHAT_DEFAULT_NUM_RESULTS,
        temperature: Optional[float] = CHAT_DEFAULT_TEMPERATURE,
        max_tokens: Optional[int] = CHAT_DEFAULT_MAX_TOKENS,
        min_tokens: Optional[int] = CHAT_DEFAULT_MIN_TOKENS,
        top_p: Optional[float] = CHAT_DEFAULT_TOP_P,
        top_k_return: Optional[int] = CHAT_DEFAULT_TOP_K_RETURN,
        stop_sequences: Optional[List[str]] | NotGiven = NOT_GIVEN,
        frequency_penalty: Optional[Penalty] | NotGiven = NOT_GIVEN,
        presence_penalty: Optional[Penalty] | NotGiven = NOT_GIVEN,
        count_penalty: Optional[Penalty] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> ChatResponse:
        if any(isinstance(item, JambaChatMessage) for item in messages):
            raise ValueError(
                "Please use the ChatMessage class from ai21.models"
                " instead of ai21.models.chat when working with chat"
            )

        body = self._create_body(
            model=model,
            messages=messages,
            system=system,
            num_results=num_results,
            temperature=temperature,
            max_tokens=max_tokens,
            min_tokens=min_tokens,
            top_p=top_p,
            top_k_return=top_k_return,
            stop_sequences=stop_sequences,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            count_penalty=count_penalty,
            **kwargs,
        )

        return self._post(path=f"/{model}/{self._module_name}", body=body, response_cls=ChatResponse)

    @property
    def completions(self) -> ChatCompletions:
        return ChatCompletions(self._client)


class AsyncStudioChat(AsyncStudioResource, Chat):
    async def create(
        self,
        model: str,
        messages: List[ChatMessage],
        system: str,
        *,
        num_results: Optional[int] = CHAT_DEFAULT_NUM_RESULTS,
        temperature: Optional[float] = CHAT_DEFAULT_TEMPERATURE,
        max_tokens: Optional[int] = CHAT_DEFAULT_MAX_TOKENS,
        min_tokens: Optional[int] = CHAT_DEFAULT_MIN_TOKENS,
        top_p: Optional[float] = CHAT_DEFAULT_TOP_P,
        top_k_return: Optional[int] = CHAT_DEFAULT_TOP_K_RETURN,
        stop_sequences: Optional[List[str]] | NotGiven = NOT_GIVEN,
        frequency_penalty: Optional[Penalty] | NotGiven = NOT_GIVEN,
        presence_penalty: Optional[Penalty] | NotGiven = NOT_GIVEN,
        count_penalty: Optional[Penalty] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> ChatResponse:
        if any(isinstance(item, JambaChatMessage) for item in messages):
            raise ValueError(
                "Please use the ChatMessage class from ai21.models"
                " instead of ai21.models.chat when working with chat"
            )

        body = self._create_body(
            model=model,
            messages=messages,
            system=system,
            num_results=num_results,
            temperature=temperature,
            max_tokens=max_tokens,
            min_tokens=min_tokens,
            top_p=top_p,
            top_k_return=top_k_return,
            stop_sequences=stop_sequences,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            count_penalty=count_penalty,
            **kwargs,
        )
        return await self._post(path=f"/{model}/{self._module_name}", body=body, response_cls=ChatResponse)

    @property
    def completions(self) -> AsyncChatCompletions:
        return AsyncChatCompletions(self._client)



from typing import Optional


class AI21APIError(Exception):
    def __init__(self, status_code: int, details: Optional[str] = None):
        super().__init__(details)
        self.details = details
        self.status_code = status_code

    def __str__(self) -> str:
        return f"Failed with http status code: {self.status_code} ({type(self).__name__}). Details: {self.details}"


class BadRequest(AI21APIError):
    def __init__(self, details: Optional[str] = None):
        super().__init__(400, details)


class Unauthorized(AI21APIError):
    def __init__(self, details: Optional[str] = None):
        super().__init__(401, details)


class AccessDenied(AI21APIError):
    def __init__(self, details: Optional[str] = None):
        super().__init__(403, details)


class NotFound(AI21APIError):
    def __init__(self, details: Optional[str] = None):
        super().__init__(404, details)


class APITimeoutError(AI21APIError):
    def __init__(self, details: Optional[str] = None):
        super().__init__(408, details)


class UnprocessableEntity(AI21APIError):
    def __init__(self, details: Optional[str] = None):
        super().__init__(422, details)


class ModelErrorException(AI21APIError):
    def __init__(self, details: Optional[str] = None):
        super().__init__(424, details)


class TooManyRequestsError(AI21APIError):
    def __init__(self, details: Optional[str] = None):
        super().__init__(429, details)


class AI21ServerError(AI21APIError):
    def __init__(self, details: Optional[str] = None):
        super().__init__(500, details)


class ServiceUnavailable(AI21APIError):
    def __init__(self, details: Optional[str] = None):
        super().__init__(500, details)


class AI21Error(Exception):
    def __init__(self, message: str):
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return f"{type(self).__name__} {self.message}"


class MissingApiKeyError(AI21Error):
    def __init__(self):
        message = "API key must be supplied either globally in the ai21 namespace, or to be provided in the call args"
        super().__init__(message)
        self.message = message


class ModelPackageDoesntExistError(AI21Error):
    def __init__(self, model_name: str, region: str, version: Optional[str] = None):
        message = f"model_name: {model_name} doesn't exist in region: {region}"

        if version is not None:
            message += f" with version: {version}"

        super().__init__(message)
        self.message = message


class EmptyMandatoryListError(AI21Error):
    def __init__(self, key: str):
        message = f"Supplied {key} is empty. At least one element should be present in the list"
        super().__init__(message)


class CredentialsError(AI21Error):
    def __init__(self, provider_name: str, error_message: str):
        message = f"Could not get default {provider_name} credentials: {error_message}"
        super().__init__(message)


class StreamingDecodeError(AI21Error):
    def __init__(self, chunk: str, error_message: Optional[str] = None):
        message = f"Failed to decode chunk: {chunk} in stream. Please check the stream format."
        if error_message:
            message = f"{message} Error: {error_message}"
        super().__init__(message)


class InternalDependencyException(AI21APIError):
    def __init__(self, details: Optional[str] = None):
        super().__init__(530, details)


class CodeParsingError(AI21Error):
    def __init__(self, details: Optional[str] = None):
        message = f"Code can't be parsed{'' if details is None else f': {details}'}"
        super().__init__(message)
        self.message = message



from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from typing import Any, Dict, Type, Callable, List

from pydantic import BaseModel

from ai21.errors import CodeParsingError
from ai21.models._pydantic_compatibility import _to_schema
from ai21.models.responses.plan_response import PlanResponse, ListPlanResponse
from ai21.types import NOT_GIVEN, NotGiven
from ai21.utils.typing import remove_not_given


class BasePlans(ABC):
    _module_name = "plans"

    def _parse_schema(self, schema: Type[BaseModel] | Dict[str, Any]) -> Dict[str, Any]:
        if inspect.isclass(schema) and issubclass(schema, BaseModel):
            return _to_schema(schema)
        return schema

    def _parse_code(self, code: str | Callable) -> str:
        if callable(code):
            try:
                return inspect.getsource(code).strip()
            except OSError as e:
                raise CodeParsingError(str(e))
            except Exception:
                raise CodeParsingError()
        return code

    @abstractmethod
    def create(
        self,
        *,
        assistant_id: str,
        code: str | Callable,
        schemas: List[Dict[str, Any]] | List[Type[BaseModel]] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> PlanResponse:
        pass

    def _create_body(
        self,
        *,
        code: str | Callable,
        schemas: List[Dict[str, Any]] | List[BaseModel] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> Dict[str, Any]:
        code_str = self._parse_code(code)

        return remove_not_given(
            {
                "code": code_str,
                "schemas": (
                    [self._parse_schema(schema) for schema in schemas] if schemas is not NOT_GIVEN else NOT_GIVEN
                ),
                **kwargs,
            }
        )

    @abstractmethod
    def list(
        self,
        *,
        assistant_id: str,
    ) -> ListPlanResponse:
        pass

    @abstractmethod
    def retrieve(
        self,
        *,
        assistant_id: str,
        plan_id: str,
    ) -> PlanResponse:
        pass

    @abstractmethod
    def modify(
        self,
        *,
        assistant_id: str,
        plan_id: str,
        code: str,
        schemas: List[Dict[str, Any]] | NotGiven = NOT_GIVEN,
    ) -> PlanResponse:
        pass



from datetime import datetime
from typing import List, Optional, Dict, Any

from ai21.models.ai21_base_model import AI21BaseModel


class PlanResponse(AI21BaseModel):
    id: str
    created_at: datetime
    updated_at: datetime
    assistant_id: str
    code: str
    schemas: List[Dict[str, Any]]


class ListPlanResponse(AI21BaseModel):
    results: List[PlanResponse]


class PlanValidationResponse(AI21BaseModel):
    is_valid: bool
    message: Optional[str] = None
    details: Optional[str] = None



from datetime import datetime
from typing import Optional

from ai21.models.ai21_base_model import AI21BaseModel
from ai21.models.assistant.assistant import Optimization
from ai21.models.assistant.run import RunStatus, RequiredAction


class RunResponse(AI21BaseModel):
    id: str
    created_at: datetime
    updated_at: datetime
    thread_id: str
    assistant_id: str
    description: Optional[str] = None
    status: RunStatus
    optimization: Optimization
    execution_id: Optional[str] = None
    required_action: Optional[RequiredAction] = None



from __future__ import annotations

from ai21.clients.common.beta.assistant.messages import BaseMessages
from ai21.clients.studio.resources.studio_resource import StudioResource, AsyncStudioResource
from ai21.models.assistant.message import ThreadMessageRole, MessageContentText
from ai21.models.responses.message_response import MessageResponse, ListMessageResponse


class ThreadMessages(StudioResource, BaseMessages):
    def create(
        self,
        thread_id: str,
        *,
        role: ThreadMessageRole,
        content: MessageContentText,
        **kwargs,
    ) -> MessageResponse:
        body = dict(
            role=role,
            content=content,
        )

        return self._post(path=f"/threads/{thread_id}/{self._module_name}", body=body, response_cls=MessageResponse)

    def list(self, thread_id: str) -> ListMessageResponse:
        return self._get(path=f"/threads/{thread_id}/{self._module_name}", response_cls=ListMessageResponse)


class AsyncThreadMessages(AsyncStudioResource, BaseMessages):
    async def create(
        self,
        thread_id: str,
        *,
        role: ThreadMessageRole,
        content: MessageContentText,
        **kwargs,
    ) -> MessageResponse:
        body = dict(
            role=role,
            content=content,
        )

        return await self._post(
            path=f"/threads/{thread_id}/{self._module_name}", body=body, response_cls=MessageResponse
        )

    async def list(self, thread_id: str) -> ListMessageResponse:
        return await self._get(path=f"/threads/{thread_id}/{self._module_name}", response_cls=ListMessageResponse)



from __future__ import annotations

import json
from functools import lru_cache
from typing import Iterator, AsyncIterator

import httpx
from botocore.eventstream import EventStreamMessage, EventStreamBuffer
from botocore.model import Shape
from botocore.parsers import EventStreamJSONParser

from ai21.errors import StreamingDecodeError
from ai21.stream.stream_commons import _SSEDecoderBase


_FINISH_REASON_NULL_STR = '"finish_reason":null'


@lru_cache(maxsize=None)
def get_response_stream_shape() -> Shape:
    from botocore.model import ServiceModel
    from botocore.loaders import Loader

    loader = Loader()
    bedrock_service_dict = loader.load_service_model("bedrock-runtime", "service-2")
    bedrock_service_model = ServiceModel(bedrock_service_dict)
    return bedrock_service_model.shape_for("ResponseStream")


class _AWSEventStreamDecoder(_SSEDecoderBase):
    def __init__(self) -> None:
        self._parser = EventStreamJSONParser()

    def iter(self, response: httpx.Response) -> Iterator[str]:
        event_stream_buffer = EventStreamBuffer()
        previous_item = None
        for chunk in response.iter_bytes():
            try:
                item = next(self._process_chunks(event_stream_buffer, chunk))
            except StopIteration as e:
                raise StreamingDecodeError(chunk=str(chunk), error_message=str(e))
            # For Bedrock metering chunk:
            if previous_item is not None:
                item = self._build_last_chunk(last_model_chunk=previous_item, bedrock_metrics_chunk=item)
            if _FINISH_REASON_NULL_STR not in item and previous_item is None:
                previous_item = item
                continue
            yield item

    async def aiter(self, response: httpx.Response) -> AsyncIterator[str]:
        event_stream_buffer = EventStreamBuffer()
        previous_item = None
        async for chunk in response.aiter_bytes():
            try:
                item = next(self._process_chunks(event_stream_buffer, chunk))
            except StopIteration as e:
                raise StreamingDecodeError(chunk=str(chunk), error_message=str(e))
            # For Bedrock metering chunk:
            if previous_item is not None:
                item = self._build_last_chunk(last_model_chunk=previous_item, bedrock_metrics_chunk=item)
            if _FINISH_REASON_NULL_STR not in item and previous_item is None:
                previous_item = item
                continue
            yield item

    def _parse_message_from_event(self, event: EventStreamMessage) -> str | None:
        response_dict = event.to_response_dict()
        parsed_response = self._parser.parse(response_dict, get_response_stream_shape())
        if response_dict["status_code"] != 200:
            raise ValueError(f"Bad response code, expected 200: {response_dict}")

        chunk = parsed_response.get("chunk")
        if not chunk:
            return None

        return chunk.get("bytes").decode()  # type: ignore[no-any-return]

    def _build_last_chunk(self, last_model_chunk: str, bedrock_metrics_chunk: str) -> str:
        chunk_dict = json.loads(last_model_chunk)
        bedrock_metrics_dict = json.loads(bedrock_metrics_chunk)
        chunk_dict = {**chunk_dict, **bedrock_metrics_dict}
        return json.dumps(chunk_dict)

    def _process_chunks(self, event_stream_buffer, chunk) -> Iterator[str]:
        try:
            event_stream_buffer.add_data(chunk)
            for event in event_stream_buffer:
                message = self._parse_message_from_event(event)
                if message:
                    yield message
        except Exception as e:
            raise StreamingDecodeError(chunk=str(chunk), error_message=str(e))



import warnings
from typing import Any, Dict

from pydantic import BaseModel, ConfigDict
from typing_extensions import Self

from ai21.models._pydantic_compatibility import _to_dict, _to_json, _from_dict, _from_json, IS_PYDANTIC_V2


class AI21BaseModel(BaseModel):
    if IS_PYDANTIC_V2:
        model_config = ConfigDict(
            populate_by_name=True,
            protected_namespaces=(),
            extra="allow",
        )
    else:

        class Config:
            from pydantic import Extra

            allow_population_by_field_name = True
            extra = Extra.allow

    def to_dict(self, **kwargs) -> Dict[str, Any]:
        warnings.warn(
            "The 'to_dict' method is deprecated and will be removed in a future version."
            " Please use Pydantic's built-in methods instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        kwargs["by_alias"] = kwargs.pop("by_alias", True)

        return _to_dict(self, **kwargs)

    def to_json(self, **kwargs) -> str:
        warnings.warn(
            "The 'to_json' method is deprecated and will be removed in a future version."
            " Please use Pydantic's built-in methods instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        kwargs["by_alias"] = kwargs.pop("by_alias", True)

        return _to_json(self, **kwargs)

    @classmethod
    def from_dict(cls, obj: Any, **kwargs) -> Self:
        warnings.warn(
            "The 'from_dict' method is deprecated and will be removed in a future version."
            " Please use Pydantic's built-in methods instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        return _from_dict(cls, obj, **kwargs)

    @classmethod
    def from_json(cls, json_str: str, **kwargs) -> Self:
        warnings.warn(
            "The 'from_json' method is deprecated and will be removed in a future version."
            " Please use Pydantic's built-in methods instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        return _from_json(cls, json_str, **kwargs)



from __future__ import annotations

from typing import List

from ai21.clients.common.beta.assistant.runs import BaseRuns
from ai21.clients.studio.resources.studio_resource import StudioResource, AsyncStudioResource
from ai21.models.assistant.assistant import Optimization
from ai21.models.assistant.run import ToolOutput
from ai21.models.responses.run_response import RunResponse
from ai21.types import NotGiven, NOT_GIVEN


class ThreadRuns(StudioResource, BaseRuns):
    def create(
        self,
        *,
        thread_id: str,
        assistant_id: str,
        description: str | NotGiven = NOT_GIVEN,
        optimization: Optimization | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> RunResponse:
        body = self._create_body(
            thread_id=thread_id,
            assistant_id=assistant_id,
            description=description,
            optimization=optimization,
            **kwargs,
        )

        return self._post(path=f"/threads/{thread_id}/{self._module_name}", body=body, response_cls=RunResponse)

    def retrieve(
        self,
        *,
        thread_id: str,
        run_id: str,
    ) -> RunResponse:
        return self._get(path=f"/threads/{thread_id}/{self._module_name}/{run_id}", response_cls=RunResponse)

    def cancel(
        self,
        *,
        thread_id: str,
        run_id: str,
    ) -> RunResponse:
        return self._post(path=f"/threads/{thread_id}/{self._module_name}/{run_id}/cancel", response_cls=RunResponse)

    def submit_tool_outputs(self, *, thread_id: str, run_id: str, tool_outputs: List[ToolOutput]) -> RunResponse:
        body = dict(tool_outputs=tool_outputs)

        return self._post(
            path=f"/threads/{thread_id}/{self._module_name}/{run_id}/submit_tool_outputs",
            body=body,
            response_cls=RunResponse,
        )


class AsyncThreadRuns(AsyncStudioResource, BaseRuns):
    async def create(
        self,
        *,
        thread_id: str,
        assistant_id: str,
        description: str | NotGiven = NOT_GIVEN,
        optimization: Optimization | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> RunResponse:
        body = self._create_body(
            thread_id=thread_id,
            assistant_id=assistant_id,
            description=description,
            optimization=optimization,
            **kwargs,
        )

        return await self._post(path=f"/threads/{thread_id}/{self._module_name}", body=body, response_cls=RunResponse)

    async def retrieve(
        self,
        *,
        thread_id: str,
        run_id: str,
    ) -> RunResponse:
        return await self._get(path=f"/threads/{thread_id}/{self._module_name}/{run_id}", response_cls=RunResponse)

    async def cancel(
        self,
        *,
        thread_id: str,
        run_id: str,
    ) -> RunResponse:
        return await self._post(
            path=f"/threads/{thread_id}/{self._module_name}/{run_id}/cancel", response_cls=RunResponse
        )

    async def submit_tool_outputs(self, *, thread_id: str, run_id: str, tool_outputs: List[ToolOutput]) -> RunResponse:
        body = dict(tool_outputs=tool_outputs)

        return await self._post(
            path=f"/threads/{thread_id}/{self._module_name}/{run_id}/submit_tool_outputs",
            body=body,
            response_cls=RunResponse,
        )



from __future__ import annotations

from typing import List, Any, Dict, Type

from pydantic import BaseModel

from ai21.clients.common.beta.assistant.plans import BasePlans
from ai21.clients.studio.resources.studio_resource import (
    AsyncStudioResource,
    StudioResource,
)
from ai21.models.responses.plan_response import PlanResponse, ListPlanResponse
from ai21.types import NotGiven, NOT_GIVEN


class AssistantPlans(StudioResource, BasePlans):
    def create(
        self,
        *,
        assistant_id: str,
        code: str,
        schemas: List[Dict[str, Any]] | List[Type[BaseModel]] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> PlanResponse:
        body = self._create_body(
            code=code,
            schemas=schemas,
            **kwargs,
        )

        return self._post(path=f"/assistants/{assistant_id}/{self._module_name}", body=body, response_cls=PlanResponse)

    def list(
        self,
        *,
        assistant_id: str,
    ) -> ListPlanResponse:
        return self._get(path=f"/assistants/{assistant_id}/{self._module_name}", response_cls=ListPlanResponse)

    def retrieve(
        self,
        *,
        assistant_id: str,
        plan_id: str,
    ) -> PlanResponse:
        return self._get(path=f"/assistants/{assistant_id}/{self._module_name}/{plan_id}", response_cls=PlanResponse)

    def modify(
        self,
        *,
        assistant_id: str,
        plan_id: str,
        code: str,
        schemas: List[Dict[str, Any]] | List[Type[BaseModel]] | NotGiven = NOT_GIVEN,
    ) -> PlanResponse:
        body = self._create_body(
            code=code,
            schemas=schemas,
        )

        return self._patch(
            path=f"/assistants/{assistant_id}/{self._module_name}/{plan_id}", body=body, response_cls=PlanResponse
        )


class AsyncAssistantPlans(AsyncStudioResource, BasePlans):
    async def create(
        self,
        *,
        assistant_id: str,
        code: str,
        schemas: List[Dict[str, Any]] | List[Type[BaseModel]] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> PlanResponse:
        body = self._create_body(
            code=code,
            schemas=schemas,
            **kwargs,
        )

        return self._post(path=f"/assistants/{assistant_id}/{self._module_name}", body=body, response_cls=PlanResponse)

    async def list(
        self,
        *,
        assistant_id: str,
    ) -> ListPlanResponse:
        return await self._get(path=f"/assistants/{assistant_id}/{self._module_name}", response_cls=ListPlanResponse)

    async def retrieve(
        self,
        *,
        assistant_id: str,
        plan_id: str,
    ) -> PlanResponse:
        return await self._get(
            path=f"/assistants/{assistant_id}/{self._module_name}/{plan_id}", response_cls=PlanResponse
        )

    async def modify(
        self,
        *,
        assistant_id: str,
        plan_id: str,
        code: str,
        schemas: List[Dict[str, Any]] | List[Type[BaseModel]] | NotGiven = NOT_GIVEN,
    ) -> PlanResponse:
        body = self._create_body(
            code=code,
            schemas=schemas,
        )

        return await self._patch(
            path=f"/assistants/{assistant_id}/{self._module_name}/{plan_id}", body=body, response_cls=PlanResponse
        )



from datetime import datetime
from typing import List

from ai21.models.ai21_base_model import AI21BaseModel


class RouteResponse(AI21BaseModel):
    id: str
    created_at: datetime
    updated_at: datetime
    assistant_id: str
    plan_id: str
    name: str
    description: str
    examples: List[str]


class ListRouteResponse(AI21BaseModel):
    results: List[RouteResponse]



from ai21.models.chat.role_type import RoleType
from ai21.models.chat_message import ChatMessage
from ai21.models.document_type import DocumentType
from ai21.models.penalty import Penalty
from ai21.models.responses.chat_response import ChatResponse, ChatOutput, FinishReason
from ai21.models.responses.completion_response import (
    CompletionsResponse,
    Completion,
    CompletionFinishReason,
    CompletionData,
    Prompt,
)
from ai21.models.responses.conversational_rag_response import ConversationalRagResponse, ConversationalRagSource
from ai21.models.responses.file_response import FileResponse

__all__ = [
    "ChatMessage",
    "RoleType",
    "Penalty",
    "DocumentType",
    "ChatResponse",
    "ChatOutput",
    "FinishReason",
    "CompletionsResponse",
    "Completion",
    "CompletionFinishReason",
    "CompletionData",
    "Prompt",
    "FileResponse",
    "ConversationalRagResponse",
    "ConversationalRagSource",
]



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, Dict, List

from ai21.models.chat import ChatMessage
from ai21.models._pydantic_compatibility import _to_dict
from ai21.models.responses.conversational_rag_response import ConversationalRagResponse
from ai21.models.retrieval_strategy import RetrievalStrategy
from ai21.types import NotGiven, NOT_GIVEN
from ai21.utils.typing import remove_not_given


class ConversationalRag(ABC):
    _module_name = "conversational-rag"

    @abstractmethod
    def create(
        self,
        messages: List[ChatMessage],
        *,
        path: str | NotGiven = NOT_GIVEN,
        labels: List[str] | NotGiven = NOT_GIVEN,
        file_ids: List[str] | NotGiven = NOT_GIVEN,
        max_segments: int | NotGiven = NOT_GIVEN,
        retrieval_strategy: RetrievalStrategy | str | NotGiven = NOT_GIVEN,
        retrieval_similarity_threshold: float | NotGiven = NOT_GIVEN,
        max_neighbors: int | NotGiven = NOT_GIVEN,
        hybrid_search_alpha: float | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> ConversationalRagResponse:
        """
        :param messages: List of ChatMessage objects.
        :param path: Search only files in the specified path or a child path.
        :param labels: Search only files with one of these labels.
        :param file_ids: List of file IDs to filter the sources by.
        :param max_segments: Maximum number of segments to retrieve.
        :param retrieval_strategy: The retrieval strategy to use.
        :param retrieval_similarity_threshold: The similarity threshold to use for retrieval.
        :param max_neighbors: Maximum number of neighbors to retrieve.
        :param hybrid_search_alpha: The alpha value to use for hybrid search.
        :param kwargs: Additional keyword arguments.
        :return: The response object.
        """
        pass

    def _create_body(
        self,
        messages: List[ChatMessage],
        *,
        path: str | NotGiven,
        labels: List[str] | NotGiven,
        file_ids: List[str] | NotGiven,
        max_segments: int | NotGiven,
        retrieval_strategy: RetrievalStrategy | str | NotGiven,
        retrieval_similarity_threshold: float | NotGiven,
        max_neighbors: int | NotGiven,
        hybrid_search_alpha: float | NotGiven,
        **kwargs,
    ) -> Dict[str, Any]:
        return remove_not_given(
            {
                "messages": [_to_dict(message) for message in messages],
                "path": path,
                "labels": labels,
                "file_ids": file_ids,
                "max_segments": max_segments,
                "retrieval_strategy": retrieval_strategy,
                "retrieval_similarity_threshold": retrieval_similarity_threshold,
                "max_neighbors": max_neighbors,
                "hybrid_search_alpha": hybrid_search_alpha,
                **kwargs,
            }
        )



from dataclasses import dataclass
from typing import Any, Dict, Optional, BinaryIO

from typing_extensions import Self


@dataclass(frozen=True)
class RequestOptions:
    url: str
    body: Dict[str, Any]
    method: str
    headers: Dict[str, Any]
    timeout: float
    path: Optional[str] = None
    params: Optional[Dict[str, Any]] = None
    stream: bool = False
    files: Optional[Dict[str, BinaryIO]] = None

    def replace(
        self,
        url: Optional[str] = None,
        body: Optional[Dict[str, Any]] = None,
        method: Optional[str] = None,
        headers: Optional[Dict[str, Any]] = None,
        timeout: Optional[float] = None,
        params: Optional[Dict[str, Any]] = None,
        stream: Optional[bool] = None,
        path: Optional[str] = None,
        files: Optional[Dict[str, BinaryIO]] = None,
    ) -> Self:
        return RequestOptions(
            url=url or self.url,
            body=body or self.body,
            method=method or self.method,
            headers=headers or self.headers,
            timeout=timeout or self.timeout,
            params=params or self.params,
            stream=stream if stream is not None else self.stream,
            path=path or self.path,
            files=files or self.files,
        )



from typing import Literal

from typing_extensions import TypedDict

ThreadMessageRole = Literal["assistant", "user"]


class MessageContentText(TypedDict):
    type: Literal["text"]
    text: str


class Message(TypedDict):
    role: ThreadMessageRole
    content: MessageContentText



from typing_extensions import Literal, List, Optional, Union, TypeAlias

from ai21.models.ai21_base_model import AI21BaseModel
from ai21.models.chat.tool_call import ToolCall


class ChatMessage(AI21BaseModel):
    role: str
    content: str


class AssistantMessage(ChatMessage):
    role: Literal["assistant"] = "assistant"
    tool_calls: Optional[List[ToolCall]] = None
    content: Optional[str] = None


class ToolMessage(ChatMessage):
    role: Literal["tool"] = "tool"
    tool_call_id: str


class UserMessage(ChatMessage):
    role: Literal["user"] = "user"


class SystemMessage(ChatMessage):
    role: Literal["system"] = "system"


ChatMessageParam: TypeAlias = Union[UserMessage, AssistantMessage, ToolMessage, SystemMessage, ChatMessage]



from typing_extensions import Literal, Any, Dict, List, TypedDict, Required


class ToolParameters(TypedDict, total=False):
    type: Literal["object"]
    properties: Required[Dict[str, Any]]
    required: List[str]



from __future__ import annotations

from abc import ABC
from typing import Optional, Callable, Dict

from ai21.clients.studio.resources.studio_chat import StudioChat, AsyncStudioChat
from ai21.http_client.async_http_client import AsyncAI21HTTPClient
from ai21.http_client.http_client import AI21HTTPClient

AzureADTokenProvider = Callable[[], str]
_DEFAULT_AZURE_VERSION = "v1"


class BaseAzureClient(ABC):
    _azure_endpoint: str
    _api_key: Optional[str]
    _azure_ad_token: Optional[str]
    _azure_ad_token_provider: Optional[AzureADTokenProvider]

    def _prepare_headers(self, headers: Dict[str, str]) -> Dict[str, str]:
        azure_ad_token = self._get_azure_ad_token()

        if azure_ad_token is not None and "Authorization" not in headers:
            return {
                "Authorization": f"Bearer {azure_ad_token}",
                **headers,
            }

        if self._api_key is not None:
            return {
                "api-key": self._api_key,
                **headers,
            }

        return headers

    def _get_azure_ad_token(self) -> Optional[str]:
        if self._azure_ad_token is not None:
            return self._azure_ad_token

        if self._azure_ad_token_provider is not None:
            return self._azure_ad_token_provider()

        return None

    def _add_version_to_url(self, base_url: str, api_version: str) -> str:
        if api_version:
            return f"{base_url}/{api_version}"

        return f"{base_url}/{_DEFAULT_AZURE_VERSION}"


class AsyncAI21AzureClient(BaseAzureClient, AsyncAI21HTTPClient):
    def __init__(
        self,
        base_url: str,
        api_version: str = _DEFAULT_AZURE_VERSION,
        api_key: Optional[str] = None,
        azure_ad_token: str | None = None,
        azure_ad_token_provider: AzureADTokenProvider | None = None,
        default_headers: Dict[str, str] | None = None,
        timeout_sec: int | None = None,
        num_retries: int | None = None,
    ):
        self._api_key = api_key
        self._azure_ad_token = azure_ad_token
        self._azure_ad_token_provider = azure_ad_token_provider

        if self._api_key is None and self._azure_ad_token_provider is None and self._azure_ad_token is None:
            raise ValueError("Must provide either api_key or azure_ad_token_provider or azure_ad_token")

        headers = self._prepare_headers(headers=default_headers or {})
        base_url = self._add_version_to_url(base_url=base_url, api_version=api_version)

        super().__init__(
            api_key=api_key,
            base_url=base_url,
            headers=headers,
            timeout_sec=timeout_sec,
            num_retries=num_retries,
            requires_api_key=False,
        )

        self.chat = AsyncStudioChat(self)
        # Override the chat.create method to match the completions endpoint,
        # so it wouldn't get to the old J2 completion endpoint
        self.chat.create = self.chat.completions.create


class AI21AzureClient(BaseAzureClient, AI21HTTPClient):
    def __init__(
        self,
        base_url: str,
        api_version: str = _DEFAULT_AZURE_VERSION,
        api_key: Optional[str] = None,
        azure_ad_token: str | None = None,
        azure_ad_token_provider: AzureADTokenProvider | None = None,
        default_headers: Dict[str, str] | None = None,
        timeout_sec: int | None = None,
        num_retries: int | None = None,
    ):
        self._api_key = api_key
        self._azure_ad_token = azure_ad_token
        self._azure_ad_token_provider = azure_ad_token_provider

        if self._api_key is None and self._azure_ad_token_provider is None and self._azure_ad_token is None:
            raise ValueError("Must provide either api_key or azure_ad_token_provider or azure_ad_token")

        headers = self._prepare_headers(headers=default_headers or {})
        base_url = self._add_version_to_url(base_url=base_url, api_version=api_version)

        super().__init__(
            api_key=api_key,
            base_url=base_url,
            headers=headers,
            timeout_sec=timeout_sec,
            num_retries=num_retries,
            requires_api_key=False,
        )

        self.chat = StudioChat(self)
        # Override the chat.create method to match the completions endpoint,
        # so it wouldn't get to the old J2 completion endpoint
        self.chat.create = self.chat.completions.create
